:- use_module(library(plunit)).
:- use_module(meta_interpreter).

:- dynamic test_info/2.

%%%%% definition of tested predicates begins

%%% predicates on lists begin

all_pos([]).
all_pos([N|L]) :- N>0,all_pos(L).

all_pos_test(L) :- solve(all_pos(L)).

member(X,[X|_]).
member(X,[Y|L]):-X\=Y,member(X,L).

member_test(X,L) :- solve(member(X,L)).

max([N],N).
max([N|L],M2) :- max(L,M),M2 is max(N,M).

max_test(L,M) :- solve(max(L,M)).

non_member(_,[]).
non_member(X,[Y|L]) :- X\=Y,non_member(X,L).

elems([],[]).
elems([X|L],S) :- elems(L,S),member(X,S).
elems([X|L],[X|S]) :- elems(L,S),non_member(X,S).

non_member_test(X,L) :- solve(non_member(X,L)).

elems_test(L,S) :- solve(elems(L,S)).

%%% predicates on lists end

%%% addition of repeating decimals starts 

add([D1|N1],[D2|N2],[D|N],C2) :- add(N1,N2,N,C), S is D1+D2+C, D is S mod 10, C2 is S div 10.   

add_test(N1,N2,N,C) :- solve(add(N1,N2,N,C)).

%%% addition of repeating decimals end 

%%% grammars begin

non_nullable(A) :- atom(A), A\==[].
non_nullable([Str1|_]) :- non_nullable(Str1).
non_nullable([_|Str2]) :- non_nullable(Str2).
non_nullable((Str1|Str2)) :- non_nullable(Str1),non_nullable(Str2).

nullable([]).
nullable([Str1|Str2]) :- nullable(Str1),nullable(Str2).
nullable((Str1|_)) :- nullable(Str1).
nullable((_|Str2)) :- nullable(Str2).

first([],[]).
first(A,[A]) :- atom(A), A\==[].
first((Str1|Str2),Set) :- first(Str1,Set1), first(Str2,Set2), ord_union(Set1,Set2,Set).
first([Str1|Str2],Set) :- nullable(Str1), first(Str1,Set1), first(Str2,Set2), ord_union(Set1,Set2,Set).
first([Str1|_],Set) :- non_nullable(Str1), first(Str1,Set). 

nullable_test(Str) :- solve(nullable(Str)).
non_nullable_test(Str) :- solve(non_nullable(Str)).
first_test(Str,Set) :- solve(first(Str,Set)).

%%% grammars end

%%% graphs begin

%%% reachable is an overloaded predicate

%%% works for a single node

reachable(node(NodeId,AdjList),NodeIdSet) :- reachable(AdjList,TempNodeIdSet), ord_add_element(TempNodeIdSet,NodeId,NodeIdSet).

%%% works for a list of nodes

reachable([],[]).

reachable([Node|NodeList],NodeIdSet) :- reachable(Node,NodeIdSet1), reachable(NodeList,NodeIdSet2), ord_union(NodeIdSet1,NodeIdSet2,NodeIdSet). 

test(Node,NodeIdSet) :- solve(reachable(Node,NodeIdSet)).

%%% graphs end

%%% paths in infinite trees begin

path0(tree(0,TL)) :- is_in(T,TL), path0(T).

is_in(T,[T|_]).
is_in(T,[_|TL]) :- is_in(T,TL).

path0_test(T) :- solve(path0(T)).

%%% paths in infinite trees end

%%% cofacts for all predicates (grouped together to avoid warnings)

cofact(all_pos(_)).

cofact(max([N|_],N)).

cofact(add(_,_,_,-1)).
cofact(add(_,_,_,0)).
cofact(add(_,_,_,1)).
cofact(add(_,_,_,2)).

cofact(non_member(_)).

cofact(elems(_,[])).

cofact(non_nullable(_)).
cofact(first(_,[])).

cofact(reachable(_,[])).

cofact(path0(_)).

%%%%% definition of tested predicates ends

settestinfo(Name,Fail) :- retractall(test_info(_,_)), assert(test_info(Name,Fail)).

:- begin_tests(solve).

test(all_pos) :- L=[3,4,3|L], all_pos_test([1,2|L]).
test(all_pos,fail) :- L=[3,4,-3|L], all_pos_test([1,2|L]).
test(member) :- L=[3,4,3|L],member_test(4,[1,2|L]).
test(member,fail) :- L=[3,4,3|L],member_test(5,[1,2|L]).
test(max1) :- L=[3,4,3|L], max_test([1,2|L],M), M==4.
test(max2) :- L=[3,4,3|L], max_test([1,5|L],M), M==5.
test(max3) :- L=[3,4,3|L], max_test([1,2|L],4).
test(max4) :- L=[3,4,3|L], max_test([1,5|L],5).
test(max1,fail) :- L=[3,4,3|L], max_test([1,2|L],0).
test(max2,fail) :- L=[3,4,3|L], max_test([1,2|L],3).
test(max3,fail) :- L=[3,4,3|L], max_test([1,2|L],5).
test(add01) :- N=[0|N],setof((M,C),add_test(N,N,M,C),L),length(L,2),Nines=[9|Nines],Zeros=[0|Zeros],memberchk((Nines,-1),L),memberchk((Zeros,0),L).
test(add02) :- N=[9|N],setof((M,C),add_test(N,N,M,C),L),length(L,2),Nines=[9|Nines],Zeros=[0|Zeros],memberchk((Nines,1),L),memberchk((Zeros,2),L). 
test(add03) :- N=[0|N],setof((M,C),add_test([1|N],[1|N],M,C),L),length(L,2),Nines=[9|Nines],Zeros=[0|Zeros],memberchk(([1|Nines],0),L),memberchk(([2|Zeros],0),L).
test(add04) :- N=[9|N],setof((M,C),add_test([1|N],[1|N],M,C),L),length(L,2),Nines=[9|Nines],Zeros=[0|Zeros],memberchk(([3|Nines],0),L),memberchk(([4|Zeros],0),L).
test(add05) :- N1=[7|N1],N2=[2|N2],setof((M,C),add_test(N1,N2,M,C),L),length(L,2),Nines=[9|Nines],Zeros=[0|Zeros],memberchk((Nines,0),L),memberchk((Zeros,1),L).
test(add06) :- N1=[7|N1],N2=[3|N2],setof((M,C),add_test(N1,N2,M,C),L),length(L,1),Ones=[1|Ones],memberchk((Ones,1),L).
test(elems) :- L=[3,4,3|L], elems_test([1,2|L],S), S==[1, 2, 3, 4].
test(nullable01,fail) :- A=([B,C]|[D,E]),B=([b,B]|[]),C=([c,C]|[c]),D=([D,d]|[]),E=([E,e]|[e]),nullable_test(A).
test(nullable01) :- A=([B,C]|[D,E]),B=([b,B]|[]),C=([c,C]|[c]),D=([D,d]|[]),E=([E,e]|[e]),non_nullable_test(A).
test(first01) :- A=([B,C]|[D,E]),B=([b,B]|[]),C=([c,C]|[c]),D=([D,d]|[]),E=([E,e]|[e]),first_test(A,Set),Set==[b,c,d,e].
test(nullable02) :- A=([B,C]|[D,E]),B=([b,B]|[]),C=([c,C]|[c]),D=([D,d]|[]),E=([E,e]|[]),nullable_test(A).
test(non_nullable01,fail) :- A=([B,C]|[D,E]),B=([b,B]|[]),C=([c,C]|[c]),D=([D,d]|[]),E=([E,e]|[]),non_nullable_test(A).
test(first02) :- A=([B,C]|[D,E]),B=([b,B]|[]),C=([c,C]|[c]),D=([D,d]|[]),E=([E,e]|[]),first_test(A,Set),Set==[b,c,d,e].
test(first03) :- A=([B,C]|[D,E]),B=([b,B]|[b]),C=([c,C]|[c]),D=([D,d]|[]),E=([E,e]|[]),first_test(A,Set),Set==[b,d,e]. 
test(first04) :- A=([B,C]|[D,E]),B=([b,B]|[b]),C=([c,C]|[c]),D=([D,d]|[d]),E=([E,e]|[]),first_test(A,Set),Set=[b,d].
test(reachable01) :- ANode = node(a,[BNode]), BNode = node(b,[ANode]), CNode = node(c,[]), solve((reachable(ANode,ASet), reachable(BNode,BSet), reachable(CNode,CSet))),length(ASet,2),length(BSet,2),length(CSet,1),memberchk(a,ASet),memberchk(a,BSet),memberchk(b,ASet),memberchk(b,BSet),memberchk(c,CSet).
test(reachable02) :- ANode = node(a,[BNode,CNode]), BNode = node(b,[ANode,CNode]), CNode = node(c,[BNode,ANode]),solve((reachable(ANode,ASet), reachable(BNode,BSet), reachable(CNode,CSet))),length(ASet,3),length(BSet,3),length(CSet,3),memberchk(a,ASet),memberchk(b,ASet),memberchk(c,ASet),memberchk(a,BSet),memberchk(b,BSet),memberchk(c,BSet),memberchk(a,CSet),memberchk(b,CSet),memberchk(c,CSet).
test(path) :- T1=tree(0,TL1), TL1=[T2,T1|TL2], T2=tree(0,TL2), TL2=[tree(1,TL1)|TL2], path0_test(T1).
test(path,fail) :- T1=tree(0,TL1), TL1=[T2,T1|TL2], T2=tree(0,TL2), TL2=[tree(1,TL1)|TL2], path0_test(T2).

:- end_tests(solve).
